import time

from data_dispatcher.api import DataDispatcherClient
from data_dispatcher import Version
from metacat.webapi import MetaCatClient 
import sys, getopt, os, json

from .cli import CLI, CLICommand, InvalidArguments
from .ui_project import ProjectCLI
from .ui_file import FileCLI
from .ui_worker import WorkerCLI
from .ui_rse import RSECLI
from .ui_lib import pretty_json

# =-=-=-=-=-=-=
# debugging hack  
# -- uncomment below to get URL's, etc. printed
#import http.client
#import logging
#http.client.HTTPConnection.debuglevel = 1
#logging.basicConfig(level=logging.DEBUG)


class LoginCommand(CLICommand):
    
    Opts = "t:m:"
    Usage = """<mechanism> <args>...             -- log in
        login -m x509 <user> <cert> [<key>]                   - use X.509 authentication
        login -m password <user>                              - password authentication
        login -m token [-t (<token>|<token file>)] <user>     - WLCG token authentication
    """
    MinArgs = 1
    
    def __call__(self, command, client, opts, args):
        
        if args[0] in ("password", "x509"):
            mech = args[0]
            args = args[1:]
        else:
            mech = opts.get("-m", "password")
        username, rest = args[0], args[1:]
        if mech == "x509":
            key = None
            if not rest:
                raise InvalidArguments("Certificate or proxy file needed")
            elif len(args) == 1:
                cert = key = rest[0]
            else:
                cert, key = rest[:2]
            user, expiration = client.login_x509(username, cert, key)
        elif mech == "password":
            import getpass
            password = getpass.getpass("Password:")
            user, expiration = client.login_password(username, password)
        elif mech == "token":
            token = token_file = None
            v = opts.get("-t")
            if v:
                if os.path.isfile(v):
                    token = open(v, "r").read().strip()
                else:
                    token = v
            else:
                token = os.environ.get("BEARER_TOKEN")
                if not token:
                    token_file = os.environ.get("BEARER_TOKEN_FILE")
                    if not token_file:
                        uid = os.environ.get("ID", str(os.geteuid()))
                        token_dir = os.environ.get("XDG_RUNTIME_DIR", "/tmp")
                        token_file = token_dir + "/" + "bt_u" + uid
            if not token:
                if token_file:
                    if not os.path.isfile(token_file):
                        print("File not found:", token_file, file=sys.stderr)
                        sys.exit(1)
                    token = open(token_file, "r").read().strip()
            if not token:
                print("Token not found", file=sys.stderr)
                sys.exit(1)
            user, expiration = client.login_token(username, token)
        else:
            raise InvalidArguments(f"Unknown authentication mechanism {mech}\n")

        if not client.tokens_saved():
            print("Authentication token not saved. Can not access/create token library", sys.stderr)
            sys.exit(1)

        print ("User:   ", user)
        print ("Expires:", time.ctime(expiration))
        

class VersionCommand(CLICommand):
    
    def __call__(self, command, client, opts, args):
        print("Server URL:    ", client.ServerURL)
        print("Server version:", client.version())
        print("Client version:", Version)

class DDCLI(CLI):
    
    Opts = "s:a:"
    Usage = """[-s <server URL>] [-a <auth server URL>]
                Both server and auth server URLs must be specified either using -s and -a or 
                via environment variables DATA_DISPATCHER_URL and DATA_DISPATCHER_AUTH_URL
            """
    
    def __init__(self):
        
        CLI.__init__(self,
            "login",    LoginCommand(),
            "version",  VersionCommand(),
            "project",  ProjectCLI,
            "file",     FileCLI,
            "worker",   WorkerCLI,
            "rse",      RSECLI
        )
            
    def update_context(self, context, command, opts, args):
        if context is None:
            server_url = opts.get("-s") or os.environ.get("DATA_DISPATCHER_URL")
            auth_server_url = opts.get("-a") or os.environ.get("DATA_DISPATCHER_AUTH_URL")

            if not server_url:
                print("Server address must be specified either using -s option or using environment variable DATA_DISPATCHER_URL", file=sys.stderr)
                sys.exit(2)

            if not auth_server_url:
                print("Authentication server address must be specified either using -a option or using environment variable DATA_DISPATCHER_AUTH_URL", file=sys.stderr)
                sys.exit(2)
    
            context = DataDispatcherClient(server_url, auth_server_url)       # return the client as context
        return context



my_name = sys.argv[0]

Usage = f"""
Data Dispatcher version: {Version}

Usage:

ddisp <command> <subcommand> <options> <args>

Commands:

    worker [-n|<worker id>]                                - set or print my worker_id 
                                                             -n generates new one
                                                             worker id will be saved in <CWD>/.worker_id
    create project ...
    show project ...
    list projects ...

    delete project <project_id>
    cancel project [-j] <project_id>                       - cancel project, -j - print project info as JSON

    show file [-j] <namespace>:<name>

    show handle [-j] <project_id> <namespace>:<name>       - show file handle, -j - as JSON

    next [-j] [-t <timeout>] [-c <cpu_site>] <project_id>  - get next available file
                                                             -c - choose the file according to the CPU/RSE proximity map for the CPU site
                                                             -j - as JSON
                                                             -t - wait for next file until "timeout" seconds, 
                                                                  otherwise, wait until the project finishes
    done <project_id> <namespace>:<name>                   - mark the file as successfully processed
    failed [-f] <project_id> <namespace>:<name>            - mark the file as failed, -f means "final", no retries

    list rses [-j]                                         - list RSEs, -j: print as JSON
    show rse [-j] <rse>                                    - show information about RSE
    set rse -a (up|down) <rse>                             - set RSE availability (requires admin privileges)
    
    login x509 <user> <cert> <key>
    login password <user>
"""


def junk():

    if command == "login":
        if len(rest) < 2:
            print(Usage)
            sys.exit(2)
        subcommand, rest = rest[0], rest[1:]
        if subcommand == "x509":
            username, cert, key = rest
            user, expiration = client.login_x509(username, cert, key)
        elif subcommand == "password":
            import getpass
            username = rest[0]
            password = getpass.getpass("Password:")
            user, expiration = client.login_password(username, password)
        else:
            print(f"Unknown authentication mechanism {subcommand}\n")
            print(Usage)
            sys.exit(2)

        print ("User:   ", user)
        print ("Expires:", time.ctime(expiration))
    
    else:
        print(Usage)
        sys.exit(2)

def main():
    cli = DDCLI()
    cli.run(sys.argv, argv0="dd")
        
if __name__ == "__main__":
    main()
        
    
            
            
        
        
        
        
        



import json
from .cli.tabular import Table, Column

def to_did(namespace, name):
    return f"{namespace}:{name}"

def from_did(did):
    return tuple(did.split(":", 1))

def pretty_json(data):
    return json.dumps(data, indent=2, sort_keys=True)
    
def parse_attrs(text):
    parts = [w.split("=",1) for w in text.split()]
    out = {}
    for k, v in parts:
        try:    v = int(v)
        except:
            try:    v = float(v)
            except:
                v = {"null":None, "true":True, "false":False}.get(v, v)
        out[k]=v
    return out
    
def print_handles(handles, print_replicas):
    
    state_order = {
        "initial":      0,
        "reserved":     1,
        "done":         3,
        "failed":       4
    }
    
    table = Table("Status", "Available", "Replicas", "Attempts", "Worker", 
            Column("File" + (" / RSE, avlbl, URL" if print_replicas else ""),
                left=True)
    )

    handles = list(handles)

    for h in handles:
        #print("print_handles: handle:", h)
        if h.get("replicas") is not None: 
            h["is_available"] = any(r["available"] and r.get("rse_available") for r in h["replicas"].values())
        else: 
            h['is_available'] = False

    handles = sorted(handles, key=lambda h: (0 if h["is_available"] else 1, state_order[h["state"]], h["attempts"], h["namespace"], h["name"]))

    for f in handles:
        rlist = f.get('replicas')
        available_replicas = 0 if rlist is None else len([r for r in rlist.values() if r["available"] and r["rse_available"]])
        nreplicas = 0 if rlist is None else len(rlist)
        state = f["state"]
        if available_replicas > 0:
            file_available = "yes"
        else:
            file_available = "no"
        table.add_row(
            state,
            file_available,
            "%4d/%-4d" % (available_replicas, nreplicas),
            f["attempts"],
            f["worker_id"],
            "%s:%s" % (f["namespace"], f["name"])
        )
        if print_replicas:
            for r in sorted(f["replicas"].values(), key=lambda r: r["preference"]):
                table.add_row(None, None, None, None, None,
                    " %-10s %-3s %s" % (r["rse"], 
                        "yes" if r["available"] else "no", r["url"] or ""
                    )
                )
    table.print()







import pytest
import os
import time

from env import env, token, auth

test_proj = "files from mengel:gen_cfg"

def test_ddisp_help(auth):
    with os.popen("ddisp help 2>&1", "r") as fin:
        data = fin.read()
        assert data.find("login") > 0
        assert data.find("version") > 0

def test_ddisp_version(auth):
    with os.popen("ddisp version", "r") as fin:
        data = fin.read()
        assert data.find("Server version") > 0
        assert data.find("Client version") > 0

def test_ddisp_login_token(auth, token):
    with os.popen(f"ddisp login -m token {os.environ['USER']}", "r") as fin:
        data = fin.read()
        assert data.find(os.environ["USER"]) > 0
        assert data.find("User") >= 0
        assert data.find("Expires") >= 0

@pytest.fixture(scope='session')
def proj_id(auth):
    with os.popen(f"ddisp project create -t 60 {test_proj} ", "r") as fin:
        data = fin.read().strip()
    return data

def test_ddisp_project_create(auth, proj_id):
    assert len(proj_id) > 0
    assert type(int(proj_id)) == int

def test_ddisp_project_show(auth, proj_id):
    with os.popen(f"ddisp project show {proj_id}", "r") as fin:
        data = fin.read()
    assert data.find("Owner") > 0
    assert data.find("Status") > 0
    assert data.find("Created") > 0

def test_ddisp_project_list(auth):
    with os.popen("ddisp project list", "r") as fin:
        data = fin.read()
    assert data.find("Owner") > 0
    assert data.find("Created") > 0
    assert data.find("State") > 0

def test_ddisp_project_list_options(auth):
    with os.popen("ddisp project list -s failed") as fin:
        data = fin.read()
    assert data.find("failed") >= 0
    assert data.find("active") < 0
    assert data.find("abandoned") < 0
    with os.popen("ddisp project list -s all -n cancelled") as fin:
        data = fin.read()
    assert data.find("active") >= 0
    assert data.find("abandoned") >= 0
    assert data.find("cancelled") < 0

def test_ddisp_file_show(auth, proj_id):
    with os.popen(f"ddisp file show {proj_id} mengel:a.fcl ", "r") as fin:
        data = fin.read()
    assert data.find(f"{proj_id}") > 0
    assert data.find("namespace") > 0
    assert data.find("state") > 0

def test_ddisp_file_list(auth, proj_id):
    with os.popen(f"ddisp file list {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("Status") > 0
    assert data.find("Replicas") > 0
    assert data.find("File") > 0

def test_ddisp_file_list_rse(auth, proj_id):
    with os.popen(f"ddisp file list -r FNAL_DCACHE_DISK_TEST {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("mengel:a.fcl") > 0
    with os.popen(f"ddisp file list -r TEST {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("mengel:a.fcl") < 0

@pytest.fixture(scope='session')
def next_file(auth, proj_id):
    with os.popen(f"ddisp worker next {proj_id} ", "r") as fin:
        did = fin.read()
    return did

def test_ddisp_worker_next(auth, next_file):
    assert len(next_file) > 0

def test_ddisp_worker_id(auth):
    with os.popen("ddisp worker id", "r") as fin:
        data = fin.read()
    assert len(data) > 0

def test_ddisp_worker_list(auth, proj_id):
    with os.popen(f"ddisp worker list {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("project_id") >= 0
    assert data.find("worker_id") >= 0
    assert data.find("namespace") >= 0

def test_ddisp_worker_list_w(auth, proj_id):
    with os.popen("ddisp worker id", "r") as fin:
        wid = fin.read().strip()
    with os.popen(f"ddisp worker list -w {wid} {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("namespace") >= 0
    assert data.find("project_id") >= 0
    assert data.find("worker_id") >= 0

def test_ddisp_project_show_state(auth, proj_id):
    with os.popen(f"ddisp project show -f initial {proj_id}", "r") as fin:
        data = fin.read()
    assert data.find(".fcl") >= 0
    with os.popen(f"ddisp project show -f done {proj_id}", "r") as fin:
        data = fin.read()
    assert len(data) == 0
    with os.popen(f"ddisp project show -f test {proj_id}", "r") as fin:
        data = fin.read()
    assert data.find("Invalid state") >= 0

# needs to be fixed
#def test_ddisp_project_search(auth, proj_id):
#    with os.popen("ddisp project search -s active", "r") as fin:
#        data = fin.read()
#    assert data.find(str(proj_id)) > 0
#    assert data.find(os.environ["USER"]) > 0

def test_ddisp_worker_done(auth, proj_id, next_file):
    with os.popen(f"ddisp worker done {proj_id} {next_file} ", "r") as fin:
        data = fin.read()
    assert len(data) == 0
    with os.popen(f"ddisp file show {proj_id} {next_file} ", "r") as fin:
        data = fin.read()
    assert data.find("done") > 0

def test_ddisp_file_list_state(auth, proj_id):
    with os.popen(f"ddisp file list -s initial {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("initial") > 0
    with os.popen(f"ddisp file list -s done {proj_id} ", "r") as fin:
        data = fin.read()
    assert data.find("done") > 0
    assert data.find("initial") < 0

def test_ddisp_worker_failed(auth, proj_id):
    # failing once causes the file to go back to the initial state to be retried
    with os.popen(f"ddisp worker next {proj_id} ", "r") as fin:
        did = fin.read()
    with os.popen(f"ddisp worker failed {proj_id} {did}", "r") as fin:
        data = fin.read()
    assert len(data) == 0
    with os.popen(f"ddisp file show {proj_id} {did}", "r") as fin:
        data = fin.read()
    assert data.find("initial") > 0

    # this time mark the file as permanently failed with the -f option
    with os.popen(f"ddisp worker next {proj_id} ", "r") as fin:
        did = fin.read()
    with os.popen(f"ddisp worker failed -f {proj_id} {did} ", "r") as fin:
        data = fin.read()
    assert len(data) == 0
    with os.popen(f"ddisp file show {proj_id} {did} ", "r") as fin:
        data = fin.read()
    assert data.find("failed") > 0

def test_ddisp_worker_failed_all(auth, proj_id):
    with os.popen(f"ddisp project show -f initial {proj_id}", "r") as fin:
        dids = fin.read()

    dids = dids.split('\n')
    dids = dids[:-1]

    for did in dids:
        os.system(f"ddisp worker next {proj_id}")

    with os.popen(f"ddisp worker failed -f {proj_id} all", "r") as fin:
        data = fin.read()
        assert len(data) == 0

    with os.popen(f"ddisp file list {proj_id}", "r") as fin:
        data = fin.read()
        assert data.find("initial") < 0
        assert data.find("failed") > 0

def test_ddisp_project_restart(auth, proj_id):
    with os.popen(f"ddisp project restart -a {proj_id}", "r") as fin:
        data = fin.read()
        assert len(data) == 0
    with os.popen(f"ddisp file list {proj_id}", "r") as fin:
        data = fin.read()
        assert data.find("initial") > 0
        assert data.find("failed") < 0

def test_ddisp_worker_done_all(auth, proj_id):
    with os.popen(f"ddisp project show -f initial {proj_id}", "r") as fin:
        dids = fin.read()
    dids = dids.split('\n')
    dids = dids[:-1]
    for did in dids:
        os.system(f"ddisp worker next {proj_id}")
    with os.popen(f"ddisp worker done {proj_id} all", "r") as fin:
        data = fin.read()
        assert len(data) ==0
    with os.popen(f"ddisp file list {proj_id}", "r") as fin:
        data = fin.read()
        assert data.find("done") > 0
        assert data.find("initial") < 0
        assert data.find("failed") < 0

@pytest.fixture(scope='session')
def proj_id_copy(auth, proj_id):
    with os.popen(f"ddisp project copy {proj_id} ", "r") as fin:
        data = fin.read().strip()
    return data

def test_ddisp_project_copy(auth, proj_id, proj_id_copy):
    assert proj_id_copy != proj_id
    assert type(int(proj_id_copy)) == int

def test_ddisp_project_copy_default_timeouts(auth, proj_id_copy):
    # check that the copied project has the same timeouts as the original project
    with os.popen(f"ddisp project show {proj_id_copy}", "r") as fin:
        data = fin.read()
    assert data.find("60.0") >= 0
    assert data.find("43200.0") >= 0

def test_ddisp_project_idle_timeout(auth, proj_id_copy):
    # check that the project is active at first
    with os.popen(f"ddisp project show {proj_id_copy} ", "r") as fin:
        data = fin.read()
    assert data.find("active") >= 0
    # check that the project is marked abandoned after timeout
    time.sleep(90)
    with os.popen(f"ddisp project show {proj_id_copy} ", "r") as fin:
        data = fin.read()
    assert data.find("abandoned") >= 0

def test_ddisp_project_activate(auth, proj_id_copy):
    os.system(f"ddisp project activate {proj_id_copy} ")
    with os.popen(f"ddisp project show {proj_id_copy} ", "r") as fin:
        data = fin.read()
    assert data.find("active") >= 0


def test_ddisp_project_cancel(auth, proj_id_copy):
    os.system(f"ddisp project cancel {proj_id_copy} ")
    with os.popen(f"ddisp project show {proj_id_copy} ", "r") as fin:
        data = fin.read()
    assert data.find("cancelled") > 0

def test_ddisp_project_delete(auth, proj_id_copy):
    os.system(f"ddisp project delete {proj_id_copy} ")
    with os.popen(f"ddisp project show {proj_id_copy} ", "r") as fin:
        data = fin.read()
    assert data.find("not found") >= 0

def test_ddisp_rse_list(auth):
    with os.popen("ddisp rse list", "r") as fin:
        data = fin.read()
    assert data.find("Name") >= 0
    assert data.find("Status") >= 0
    assert data.find("Description") >= 0

#def test_ddisp_rse_set():


def test_ddisp_rse_show(auth):
    with os.popen("ddisp rse show FNAL_DCACHE_DISK_TEST", "r") as fin:
        data = fin.read()
    assert data.find("RSE") >= 0
    assert data.find("Available") >= 0






import pytest
import os
import time

from env import env, token, auth
from test_commandline import proj_id

test_proj = "files from mengel:gen_cfg"

def test_issue_18(proj_id):
    # reserve a file as worker_id a, 
    # then try to release it as worker_id b, should fail
    # then try to release it as worker_id a, should succeed

    with os.popen("ddisp worker id -n", "r") as fin:
        wid1 = fin.read().strip()
    with os.popen("ddisp worker id -n", "r") as fin:
        wid2 = fin.read().strip()
    with os.popen(f"ddisp worker id {wid1}", "r") as fin:
        wid = fin.read().strip()
    with os.popen(f"ddisp worker next {proj_id} ", "r") as fin:
        reserved_file = fin.read().strip()
    # so now we've reserved a file
    # try to mark it done with a different worker id:
    with os.popen(f"ddisp worker id {wid2}", "r") as fin:
        wid = fin.read().strip()
    with os.popen(f"ddisp worker done {proj_id} {reserved_file} ", "r") as fin:
        out = fin.read()
    assert(out.find("wrong worker_id") >= 0)

    # now mark it done with the original worker id:
    with os.popen(f"ddisp worker id {wid1}", "r") as fin:
        wid = fin.read().strip()
    with os.popen(f"ddisp worker done {proj_id} {reserved_file} ", "r") as fin:
        out = fin.read()
    assert(len(out.strip()) == 0)



import sys
from .ui_lib import to_did, from_did, pretty_json, print_handles
from .cli import CLI, CLICommand, InvalidOptions, InvalidArguments

class ShowCommand(CLICommand):
    
    Opts = "j"
    Usage = """[-j] <project_id> <file DID>
        -j                  -- JSON output
    """
    MinArgs = 2

    def show_handle(self, client, project_id, did, opts):
        namespace, name = from_did(did)
        handle = client.get_handle(project_id, namespace, name)
        if not handle:
            print(f"Handle {handle_id} not found")
            sys.exit(1)
        if "-j" in opts:
            print(pretty_json(handle))
        else:
            for name, val in handle.items():
                if name != "replicas":
                    print(f"{name:10s}: {val}")
            if "replicas" in handle:
                print("replicas:")
                replicas = handle["replicas"]
                for rse, r in replicas.items():
                    r["rse"] = rse
                replicas = sorted(replicas.values(), key=lambda r: (-r["preference"], r["rse"]))
                for r in replicas:
                    print("  Preference: ", r["preference"])
                    print("  RSE:        ", r["rse"])
                    print("  Path:       ", r["path"] or "")
                    print("  URL:        ", r["url"] or "")
                    print("  URLs:       ", r["urls"] or "")
                    print("  Available:  ", "yes" if r["available"] else "no")

    def __call__(self, command, client, opts, args):
        did = args[1]
        project_id = int(args[0])
        return self.show_handle(client, project_id, did, opts)

class ListHandlesCommand(CLICommand):
    Opts = "js:r:"
    Usage = """[options] <project id>
        -j                  -- JSON output
        -s <handle state>   -- list handles in state
        -r <rse>            -- list handles with replicas in RSE
    """
    MinArgs = 1

    def __call__(self, command, client, opts, args):
        project_id = int(args[0])
        proj = client.get_project(project_id, with_files=True, with_replicas=True)
        lst = proj["file_handles"]
        if "-s" in opts:
            filter_state = opts["-s"]
            filtered_state=[]
            for h in lst:
                state = h["state"]
                if filter_state==state:
                    filtered_state.append(h)
        else:
            filtered_state=lst

        if "-r" in opts:
            filter_rse = opts["-r"]
            filtered_list=[]
            for h in filtered_state:
                if filter_rse in h['replicas']:
                    filtered_list.append(h)
        else:
            filtered_list=filtered_state

        if "-j" in opts:
            print(pretty_json(filtered_list))
        else:
            print_handles(filtered_list, print_replicas=False)


FileCLI = CLI(
    "show",    ShowCommand(),
    "list", ListHandlesCommand()
)



from lark import Lark
from .grammar import ProjectQueryGrammar
from metacat.common.trees import Converter, Node
from metacat.common import MetaExpressionDNF
from metacat.util import insert_sql
from data_dispatcher.db import DBProject

class ProjectQueryConverter(Converter):
    
    def __default__(self, typ, children, meta):
        return Node(typ, children, _meta=meta)

    def int_constant(self, args):
        v = args[0]
        return Node("int", value=int(v.value))
        
    def float_constant(self, args):
        v = args[0]
        return Node("float", value=float(v.value))

    def bool_constant(self, args):
        v = args[0]
        #print("bool_constant:", args, args[0].value)
        return Node("bool", value=v.value.lower() == "true")

    def string_constant(self, args):
        v = args[0]
        s = v.value
        if s[0] in ('"', "'"):
            s = s[1:-1]
        if '"' in s or "'" in s:        # sanitize
            raise ValueError("Unsafe string constant containing double or single quote: %s" % (repr(s),))
        return Node("string", value=s)

    def constant_list(self, args):
        return [n["value"] for n in args]

    def scalar(self, args):
        (t,) = args
        if isinstance(t, Node) and t.T == "attribute":
            return t
        return Node("scalar", name=t.value)

    def array_any(self, args):
        (n,) = args
        return Node("array_any", name=n.value)
        
    def array_all(self, args):
        (n,) = args
        return Node("array_all", name=n.value)
        
    def array_length(self, args):
        (n,) = args
        return Node("array_length", name=n.value)
        
    def array_subscript(self, args):
        name, inx = args
        if inx.type == "STRING":
            inx = inx.value[1:-1]
        else:
            inx = int(inx.value)
        return Node("array_subscript", name=name.value, index=inx)

    def cmp_op(self, args):
        node = Node("cmp_op", [args[0], args[2]], op=args[1].value, neg=False)
        return self._convert_array_all(node)
        
    def constant_in_array(self, args):
        return Node("cmp_op",
            [Node("array_any", name=args[1].value), args[0]], op="=", neg=False
        )
        
    def constant_not_in_array(self, args):
        return Node("cmp_op",
            [Node("array_any", name=args[1].value), args[0]], op="=", neg=True
        )

    def constant_in_attr(self, args):
        const_arg = args[0]
        const_type = const_arg.T
        const_value = const_arg["value"]
        assert const_type == "string"
        return Node("cmp_op", 
                    [
                        Node("object_attribute", name=args[1].value), 
                        Node("string", value=".*%s.*" % (const_value,))
                    ], op="~", neg=False
        )


    def constant_not_in_attr(self, args):
        out = self.constant_in_attr(args)
        return out.clone(neg = not out["neg"])

    def constant_in_meta(self, args):
        const_arg = args[0]
        const_type = const_arg.T
        const_value = const_arg["value"]
        array_in = Node("cmp_op", [Node("array_any", name=args[1].value), const_arg], op="=", neg=False)
        if const_type == "string":
            return Node("meta_or",
                [   array_in,
                    Node("cmp_op", [
                        Node("meta_attribute", name=args[1].value), 
                        Node("string", value=".*%s.*" % (const_value,))
                    ], op="~", neg=False)
                ]
            )
        else:
            return array_in
        
    def constant_not_in_meta(self, args):
        const_arg = args[0]
        const_type = const_arg.T
        const_value = const_arg["value"]
        array_not_in = Node("cmp_op", [Node("array_any", name=args[1].value), const_arg], op="=", neg=True)
        if const_type == "string":
            return Node("meta_and",
                [   array_not_in,
                    Node("cmp_op", [
                        Node("meta_attribute", name=args[1].value), 
                        Node("string", value=".*%s.*" % (const_value,))
                    ], op="~", neg=True)
                ]
            )
        else:
            return array_not_in

    def in_range(self, args):
        assert len(args) == 3 and args[1].T in ("string", "int", "float") and args[2].T in ("string", "int", "float")
        assert args[1].T == args[2].T, "Range ends must be of the same type"
        return self._convert_array_all(Node("in_range", [args[0]], low=args[1]["value"], high=args[2]["value"], neg=False, type=args[1].T))
    
    def not_in_range(self, args):
        assert len(args) == 3 and args[1].T in ("string", "int", "float") and args[2].T in ("string", "int", "float")
        assert args[1].T == args[2].T, "Range ends must be of the same type"
        return self._convert_array_all(Node("in_range", [args[0]], low=args[1]["value"], high=args[2]["value"], neg=True, type=args[1].T))

    def in_set(self, args):
        assert len(args) == 2
        return self._convert_array_all(Node("in_set", [args[0]], neg=False, set=args[1]))
        
    def not_in_set(self, args):
        assert len(args) == 2
        return self._convert_array_all(Node("in_set", [args[0]], neg=True, set=args[1]))
        
    def index(self, args):
        return args[0].value
        
    def meta_and(self, args):
        children = []
        for a in args:
            if a.T == "meta_and":
                children += a.C
            else:
                children.append(a)
        return Node("meta_and", children)
        
    def meta_or(self, args):
        children = []
        for a in args:
            if a.T == "meta_or":
                children += a.C
            else:
                children.append(a)
        return Node("meta_or", children)
        
    def present(self, args):
        assert len(args) == 1
        return Node("present", name = args[0].value)

    def not_present(self, args):
        assert len(args) == 1
        return Node("not_present", name = args[0].value)

    def _apply_not(self, node):
        
        def reverse_array_wildcard(node):
            if node.T == "array_any":
                node = node.clone()
                node.T = "array_all"
            elif node.T == "array_all":
                node = node.clone()
                node.T = "node_any"
            else:
                pass
            return node
        
        if node.T in ("meta_and", "meta_or") and len(node.C) == 1:
            return self._apply_not(node.C[0])
        if node.T == "meta_and":
            return Node("meta_or", [self._apply_not(c) for c in node.C])
        elif node.T == "meta_or":
            return Node("meta_and", [self._apply_not(c) for c in node.C])
        elif node.T == "meta_not":
            return node.C[0]
        elif node.T in ("cmp_op", "in_set", "in_range"):
            node["neg"] = not node["neg"]
            return node
        elif node.T == "cmp_op":
            new_op = {
                "~":   "!~",
                "!~":  "~",
                "~*":   "!~*",
                "!~*":  "~*",
                ">":    "<=",
                "<":    ">=",
                ">=":    "<",
                "<=":    ">",
                "=":    "!=",
                "==":    "!=",
                "!=":    "=="
            }[node["op"]]
            return node.clone(op=new_op)
        elif node.T == "present":
            return Node("not_present", name=node["name"])
        elif node.T == "not_present":
            return Node("present", name=node["name"])
        else:
            raise ValueError("Unknown node type %s while trying to apply NOT operation" % (node.T,))
            
    def meta_not(self, children):
        assert len(children) == 1
        return self._apply_not(children[0])
        
    def meta_attribute(self, args):
        assert len(args) == 1
        word = args[0].value
        return Node("meta_attribute", name=word)

    def object_attribute(self, args):
        assert len(args) == 1
        word = args[0].value
        assert word in ("owner", "state", "created", "ended", "id", "query")
        return Node("object_attribute", name=word)

    def _convert_array_all(self, node):
        left = node.C[0]
        if left.T == "array_all":
            if node.T == "cmp_op":
                new_op = {
                    "~":   "!~",
                    "!~":  "~",
                    "~*":   "!~*",
                    "!~*":  "~*",
                    ">":    "<=",
                    "<":    ">=",
                    ">=":    "<",
                    "<=":    ">",
                    "=":    "!=",
                    "==":    "!=",
                    "!=":    "=="
                }[node["op"]]
                node["op"] = new_op
            else:
                node.T = {
                    "in_set":"not_in_set",
                    "in_range":"not_in_range",
                    "not_in_set":"in_set",
                    "not_in_range":"in_range",
                }[node.T]
            left.T = "array_any"
            node["neg"] = not node["neg"]
        #print("_convert_array_all: returning:", node.pretty())
        return node

class ProjectQuery(object):
    
    QueryParser = Lark(ProjectQueryGrammar, start="metadata_expression")

    def __init__(self, text):
        self.Text = text
        self.Parsed = self.Converted = None
        
    def parse(self):
        self.Parsed = self.QueryParser.parse(self.Text)
        return self.Parsed
        
    def convert(self):
        self.Converted = ProjectQueryConverter()(self.parse())
        #print("converted:", self.Converted.pretty())
        return self.Converted
        
    def sql(self):
        table = DBProject.Table
        columns = DBProject.columns(table)
        meta_sql = MetaExpressionDNF(self.convert()).sql(table, "attributes")
        return insert_sql(f"""
            select {columns} 
                from {table} --
                where 
                    $meta_sql
        """, meta_sql=meta_sql)



import sys, time, json
from .ui_lib import to_did, from_did, pretty_json
from .cli import CLI, CLICommand, InvalidOptions, InvalidArguments
from data_dispatcher.api import NotFoundError

class NextFileCommand(CLICommand):
    
    Opts = "jt:c:w:"
    MinArgs = 1
    Usage = """[options] <project_id> -- get next available file
             -w <worker id>     -- specify worker id
             -c <cpu site>      -- choose the file according to the CPU/RSE proximity map for the CPU site
             -j <json file>     -- write reserved file information into a JSON file
             -t <timeout>       -- wait for next file until "timeout" seconds, 
                                   otherwise, wait until the project finishes
    """

    def __call__(self, command, client, opts, args):
        project_id = int(args[0])
        worker_id = opts.get("-w")
        json_out = opts.get("-j")
        timeout = opts.get("-t")
        if timeout is not None: timeout = int(timeout)
        cpu_site = opts.get("-c")

        try:
            reply = client.next_file(project_id, cpu_site=cpu_site, worker_id=worker_id, timeout=timeout)
        except NotFoundError:
            print("project not found")
            sys.exit(1)

        if isinstance(reply, dict):
            print("%s:%s" % (reply["namespace"], reply["name"]))
            if json_out:
                reply["replicas"] = sorted(reply["replicas"].values(), key=lambda r: 1000000 if r.get("preference") is None else r["preference"])
                with open(json_out, "w") as jo:
                    json.dump(reply, jo, indent=4, sort_keys=True)
        else:
            print("timeout" if reply else "done")
            sys.exit(1)        # timeout
           

class DoneCommand(CLICommand):
    
    MinArgs = 2
    Usage = """<project id> (<DID>|all)                          -- mark a file as done
        "all" means mark all files reserved by the worker as done
    """

    def __call__(self, command, client, opts, args):
        project_id, did = args
        if did == "all":
            dids = [to_did(h["namespace"], h["name"]) for h in client.reserved_handles(project_id)]
        else:
            dids = [did]
        for did in dids:
            client.file_done(int(project_id), did)
    
class FailedCommand(CLICommand):
    
    Opts = "f"
    MinArgs = 2
    Usage = """[-f] <project id> (<DID>|all)                      -- mark a file as failed
        -f            -- final, do not retry the file
        "all" means mark all files reserved by the worker as failed
    """
    
    def __call__(self, command, client, opts, args):
        project_id, did = args
        if did == "all":
            dids = [to_did(h["namespace"], h["name"]) for h in client.reserved_handles(project_id)]
        else:
            dids = [did]
        for did in dids:
            client.file_failed(int(project_id), did, retry = not "-f" in opts)

class IDCommand(CLICommand):
    
    Opts = "n"
    Usage = """[-n|<worker id>]                                 -- set or print worker id
        -n          -- generate random worker id
        
        worker id will be saved in <CWD>/.worker_id
    """
    
    def __call__(self, command, client, opts, args):
        if "-n" in opts:
            worker_id = client.new_worker_id()
        elif args:
            worker_id = args[0]
            client.new_worker_id(worker_id)
        else:
            worker_id = client.WorkerID
        if not worker_id:
            print("worker id unknown")
            sys.exit(1)
        print(worker_id)
        
class ListReservedCommand(CLICommand):
    
    MinArgs = 1
    Opts = "jw:"
    Usage = """[-j] [-w <worker id>] <project id>              -- list files allocated to the worker
        -j                      -- as JSON
        -w <worker id>          -- specify worker id. Otherwise, use my worker id    
    """

    def __call__(self, command, client, opts, args):
        project_id = int(args[0])
        worker_id = opts.get("-w", client.WorkerID)
        as_json = "-j" in opts
        
        try:    handles = client.reserved_handles(project_id, worker_id)
        except NotFoundError:
            print("project not found", file=sys.stderr)
            sys.exit(1)

        if as_json:
            print(pretty_json(handles))
        else:
            for h in handles:
                for name, val in h.items():
                    print(f"{name:10s}: {val}")
                


WorkerCLI = CLI(
    "id",       IDCommand(),
    "list",     ListReservedCommand(),
    "next",     NextFileCommand(),
    "done",     DoneCommand(),
    "failed",   FailedCommand()
)
